from sanic import Sanic, text
from tortoise.exceptions import IntegrityError

from sanic_security.authentication import register, login, requires_authentication
from sanic_security.authorization import require_roles, require_permissions
from sanic_security.blueprints import security
from sanic_security.captcha import request_captcha, requires_captcha
from sanic_security.exceptions import SecurityError
from sanic_security.lib.tortoise import initialize_security_orm
from sanic_security.models import Account, Permission, Role
from sanic_security.utils import json, hash_password
from sanic_security.verification import (
    request_two_step_verification,
    requires_two_step_verification,
)

app = Sanic(__name__)


@app.post("api/test/auth/register")
async def on_register(request):
    """
    Register an account with an email, username, and password. Once the account is created successfully, a two-step session is requested and the code is provided in the response.
    """
    two_step_session = await register(request)
    response = json("Registration successful!", two_step_session.code)
    two_step_session.encode(response, secure=False)
    return response


@app.post("api/test/auth/login")
async def on_login(request):
    """
    Login with an email and password. Will only encode an authentication session when logging in with the test account.
    """
    authentication_session = await login(request)
    response = json("Login successful!", authentication_session.json())
    authentication_session.encode(response, False)
    return response


@app.post("api/test/capt/request")
async def on_captcha_request(request):
    """
    Requests new captcha session.
    """
    captcha_session = await request_captcha(request)
    response = json("Captcha request successful!", captcha_session.code)
    captcha_session.encode(response, False)
    return response


@app.post("api/test/capt/attempt")
@requires_captcha()
async def on_captcha_attempt(request, captcha_session):
    """
    Captcha challenge attempt using the captcha provided in the captcha request response.
    """
    return json("Captcha attempt successful!", captcha_session.json())


@app.post("api/test/verif/request")
async def on_request_verification(request):
    """
    Requests new two-step session.
    """
    two_step_session = await request_two_step_verification(request)
    response = json("Verification request successful!", two_step_session.code)
    two_step_session.encode(response, False)
    return response


@app.post("api/test/verif/attempt")
@requires_two_step_verification()
async def on_verification_attempt(request, two_step_session):
    """
    Two-step verification attempt using the captcha provided in the verification request response.
    """
    return json("Two step verification attempt successful!", two_step_session.json())


@app.post("api/test/auth/perms/assign")
@requires_authentication()
async def on_perms_assignment(request, authentication_session):
    """
    Assigns permissions to a logged in test account.
    """
    if await Permission.filter(account=authentication_session.account).exists():
        assignment_response = json(
            "Permissions already added to account!",
            authentication_session.account.json(),
        )
    else:
        await Permission.create(
            account=authentication_session.account, wildcard="admin:update"
        )
        await Permission.create(
            account=authentication_session.account, wildcard="admin:add"
        )
        assignment_response = json(
            "Permissions added to account!", authentication_session.account.json()
        )
    return assignment_response


@app.post("api/test/auth/roles/assign")
@requires_authentication()
async def on_roles_assignment(request, authentication_session):
    """
    Assigns roles to a logged in test account account.
    """
    if await Role.filter(account=authentication_session.account).exists():
        assignment_response = json(
            "Roles already added to account!", authentication_session.account.json()
        )
    else:
        await Role.create(account=authentication_session.account, name="Admin")
        await Role.create(account=authentication_session.account, name="Mod")
        assignment_response = json(
            "Roles added to account!", authentication_session.account.json()
        )
    return assignment_response


@app.post("api/test/auth/perms/permit")
@require_permissions("admin:update")
async def on_permission_authorization_permit_attempt(request, authentication_session):
    """
    Authorization with permissions provided to the test account.
    """
    return text("Account permitted.")


@app.post("api/test/auth/roles/permit")
@require_roles("Admin", "Mod")
async def on_role_authorization_permit_attempt(request, authentication_session):
    """
    Authorization with roles provided to the test account.
    """
    return text("Account permitted.")


@app.post("api/test/account/create")
async def on_account_creation(request):
    """
    Creates an account to be used for testing purposes.
    """
    form = request.form
    try:
        account = await Account.create(
            username="test",
            email=form.get("email"),
            password=hash_password("testtest"),
            verified=True,
        )
        response = json("Account creation successful!", account.json())
    except IntegrityError:
        response = json(
            "Account creation has failed due to an expected integrity error!", None
        )
    return response


@app.exception(SecurityError)
async def on_error(request, exception):
    return exception.response


initialize_security_orm(app)
app.blueprint(security)
if __name__ == "__main__":
    app.run(host="127.0.0.1", port=8000, debug=True, workers=4)



import functools

from sanic.request import Request

from sanic_security.exceptions import UnverifiedError
from sanic_security.models import (
    Account,
    TwoStepSession,
    SessionFactory,
)
from sanic_security.validation import validate_account, validate_session

session_factory = SessionFactory()


def _validate_account(account: Account, allow_unverified: bool):
    """
    Validates that an account used for verification does not have any error conditions and also will bypass any unverified errors
    if allow_unverified is true.

    Args:
        account (Account): The account being validated.
        allow_unverified (bool): Prevents an account unverified error from raising when true, best used for registration cases.

    Raises:
        AccountError
    """
    try:
        validate_account(account)
    except UnverifiedError as e:
        if not allow_unverified:
            raise e


async def request_two_step_verification(
    request: Request, account=None, allow_unverified=False
):
    """
    Creates a two-step session associated with an account.

    Args:
        request (Request): Sanic request parameter. All request bodies are sent as form-data with the following arguments: email.
        account (Account): The account being associated with the verification session. If None, an account is retrieved via email with the form-data argument.
        allow_unverified (bool): Prevents an unverified account from raising an Unverified error.

    Returns:
         two_step_session
    """
    if not account:
        account = await Account.get_via_email(request.form.get("email"))
    _validate_account(account, allow_unverified)
    return await session_factory.get("twostep", request, account=account)


async def verify_account(two_step_session: TwoStepSession):
    """
    Used to verify an account associated to an existing two-step session.

    Args:
        two_step_session (TwoStepSession): Two-step session containing account being verified.

    Raises:
        SessionError

    Returns:
         two_step_session
    """
    two_step_session.account.verified = True
    await two_step_session.account.save(update_fields=["verified"])
    return two_step_session


async def two_step_verification(request: Request, allow_unverified=False):
    """
    Verifies a two-step verification attempt.

    Args:
        request (Request): Sanic request parameter. All request bodies are sent as form-data with the following arguments: code.
        allow_unverified (bool): Prevents an unverified account from raising an Unverified error.

    Raises:
        SessionError
        AccountError

    Returns:
         two_step_session
    """
    two_step_session = await TwoStepSession().decode(request)
    _validate_account(two_step_session.account, allow_unverified)
    validate_session(two_step_session)
    await two_step_session.crosscheck_location(request)
    await two_step_session.crosscheck_code(request.form.get("code"))
    return two_step_session


def requires_two_step_verification(allow_unverified=False):
    """
    Verifies a two-step challenge attempt.

    Args:
        allow_unverified (bool): Prevents an account unverified error from raising when true, best used for registration cases.


    Example:
        This method is not called directly and instead used as a decorator:

            @app.post("api/verification/attempt")
            @requires_two_step_verification()
            async def on_verified(request, two_step_session):
                response = json("Two-step verification attempt successful!", two_step_session.json())
                return response

    Raises:
        SessionError
        AccountError
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            two_step_session = await two_step_verification(request, allow_unverified)
            return await func(request, two_step_session, *args, **kwargs)

        return wrapped

    return wrapper



import json

import httpx

from sanic_security.utils import config


class TwillioError(Exception):
    pass


async def send_sms(to, msg):
    """
    Sends a text directly to a phone number using Twilio.

    Args:
        to (str): Phone number receiving a text, for example: 12092819472
        msg (str): Message being sent in a text.
    """
    account_sid = config["TWILIO"]["sid"]
    auth_token = config["TWILIO"]["token"]
    from_num = f"+{config['TWILIO']['from']}"
    if account_sid and auth_token and from_num:
        auth = httpx.BasicAuth(username=account_sid, password=auth_token)
        async with httpx.AsyncClient(auth=auth) as session:
            post = await session.post(
                f"https://api.twilio.com/2010-04-01/Accounts/{account_sid}/Messages.json",
                data={"From": from_num, "To": f"+{to}", "Body": msg},
            )
            if post.status_code != 201:
                raise TwillioError(json.loads(post.content)["message"])
    else:
        raise RuntimeWarning("Twilio credentials not found.")



from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from aiosmtplib import send

from sanic_security.utils import config


async def send_email(to, subj, msg, text_type="plain"):
    """
    Sends an email using SMTP.

    Args:
        to (str): Email being sent too address.
        subj (str): Email subject.
        msg (str): Email body.
        text_type (str): Can be html or plain.
    """
    message = MIMEMultipart("alternative")
    message["From"] = config["SMTP"]["from"]
    message["To"] = to
    message["Subject"] = subj
    mime_text = MIMEText(msg, text_type, "utf-8")
    message.attach(mime_text)
    await send(
        message,
        hostname=config["SMTP"]["host"],
        port=int(config["SMTP"]["port"]),
        username=config["SMTP"]["username"],
        password=config["SMTP"]["password"],
        use_tls=config["SMTP"]["tls"] == "true",
        start_tls=config["SMTP"]["start_tls"] == "true",
    )



import functools
import re

from sanic.request import Request
from tortoise.exceptions import IntegrityError, ValidationError

from sanic_security.exceptions import (
    PasswordIncorrectError,
    ExistsError,
    InvalidIdentifierError,
    NotFoundError,
)
from sanic_security.models import (
    Account,
    SessionFactory,
    AuthenticationSession,
)
from sanic_security.utils import hash_password
from sanic_security.validation import validate_account, validate_session
from sanic_security.verification import request_two_step_verification

session_factory = SessionFactory()


async def register(request: Request, verified: bool = False, disabled: bool = False):
    """
    Registers a new account to be used by a client.

    Args:
        request (Request): Sanic request parameter. All request bodies are sent as form-data with the following arguments: email, username, password, phone (including country code).
        verified (bool): If false, account being registered must be verified before use.
        disabled (bool): If true, account being registered must be enabled before use.

    Returns:
        account: An account is returned if the verified parameter is true.
        two_step_session: A two-step session is returned if the verified parameter is false.

    Raises:
        AccountError
    """
    form = request.form
    if not re.search("[^@]+@[^@]+.[^@]+", form.get("email")):
        raise InvalidIdentifierError(
            "Please use a valid email format such as you@mail.com."
        )
    if form.get("phone") and (
        not form.get("phone").isdigit() or len(form.get("phone")) < 11
    ):
        raise InvalidIdentifierError(
            "Please use a valid phone format such as 15621435489 or 19498963648018."
        )
    try:
        account = await Account.create(
            email=form.get("email"),
            username=form.get("username"),
            password=hash_password(form.get("password")),
            phone=form.get("phone"),
            verified=verified,
            disabled=disabled,
        )
        return (
            await request_two_step_verification(request, account, True)
            if not verified
            else account
        )
    except IntegrityError as ie:
        if ie.args[0].args[0] == 1062:
            raise ExistsError()
        else:
            raise ie
    except ValidationError:
        raise InvalidIdentifierError(
            "Email, username, or phone number is too long or invalid."
        )


async def login(request: Request, account: Account = None):
    """
    Used to login to accounts registered with Sanic Security.

    Args:
        request (Request): Sanic request parameter. All request bodies are sent as form-data with the following arguments: email, password.
        account (Account): Account being logged into.

    Returns:
        authentication_session

    Raises:
        AccountError
        SessionError
    """
    form = request.form
    if not account:
        account = await Account.get_via_email(form.get("email"))
    if account:
        if account.password == hash_password(form.get("password")):
            validate_account(account)
            authentication_session = await session_factory.get(
                "authentication", request, account=account
            )
            return authentication_session
        else:
            raise PasswordIncorrectError()
    else:
        raise NotFoundError("An account with this email does not exist.")


async def logout(authentication_session: AuthenticationSession):
    """
    Invalidates client's authentication session and revokes access.

    Args:
        authentication_session (AuthenticationSession): Authentication session being invalidated and logged out from.
    """
    authentication_session.valid = False
    await authentication_session.save(update_fields=["valid"])


async def authenticate(request: Request):
    """
    Used to determine if the client is authenticated.

    Args:
        request (Request): Sanic request parameter.

    Returns:
        authentication_session

    Raises:
        AccountError
        SessionError
    """
    authentication_session = await AuthenticationSession().decode(request)
    validate_account(authentication_session.account)
    validate_session(authentication_session)
    await authentication_session.crosscheck_location(request)
    return authentication_session


def requires_authentication():
    """
    Used to determine if the client is authenticated.

    Example:
        This method is not called directly and instead used as a decorator:

            @app.post('api/authenticate')
            @requires_authentication()
            async def on_authenticate(request, authentication_session):
                return text('User is authenticated!')

    Raises:
        AccountError
        SessionError
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            authentication_session = await authenticate(request)
            return await func(request, authentication_session, *args, **kwargs)

        return wrapped

    return wrapper



import setuptools
from setuptools import setup

with open("README.md", "r") as rm:
    long_description = rm.read()

setup(
    name="sanic-security",
    version="0.10.4.1",
    packages=setuptools.find_packages(),
    url="https://github.com/sunset-developer/sanic-security",
    license="GNU General Public License v3.0",
    author="sunset-developer",
    install_requires=[
        "sanic>=21.3.0",
        "tortoise-orm>=0.17.0",
        "aiofiles",
        "pyjwt",
        "captcha",
        "aiosmtplib",
        "httpx"
    ],
    author_email="aidanstewart@sunsetdeveloper.com",
    description="A powerful, simple, and async security library for Sanic.",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



import datetime

from sanic_security.exceptions import (
    NotFoundError,
    DeletedError,
    DisabledError,
    UnverifiedError,
    InvalidError,
    ExpiredError,
)
from sanic_security.models import Account, Session


def validate_account(account: Account):
    """
    Validates an account by determining if an error should be raised due to variable values.

    Args:
        account (Account): Account being validated.

    Raises:
        AccountError
    """
    if not account:
        raise NotFoundError("This account does not exist.")
    elif account.deleted:
        raise DeletedError("This account has been permanently deleted.")
    elif account.disabled:
        raise DisabledError()
    elif not account.verified:
        raise UnverifiedError()


def validate_session(session: Session):
    """
    Validates a session by determining if an error should be raised due to variable values.

    Args:
        session (Session): Session being validated.

    Raises:
        SessionError
    """
    if session is None:
        raise NotFoundError("Session could not be found.")
    elif not session.valid:
        raise InvalidError()
    elif session.deleted:
        raise DeletedError("Session has been deleted.")
    elif datetime.datetime.now(datetime.timezone.utc) >= session.expiration_date:
        raise ExpiredError()
